import logging
import warnings
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
from typing import Dict, Any, List, Union, Tuple
import pandas as pd
import scipy.stats
from numba import jit
from scipy import sparse
from scipy.spatial.distance import pdist, squareform
from scipy.stats import norm as normal
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.neighbors import NearestNeighbors

from velocyto.diffusion import Diffusion
from velocyto.estimation import (colDeltaCor, colDeltaCorLog10,
                                 colDeltaCorLog10partial, colDeltaCorpartial,
                                 colDeltaCorSqrt, colDeltaCorSqrtpartial)
#from velocyto.neighbors import (BalancedKNN, connectivity_to_weights,
#                                convolve_by_sparse_weights,
#                                knn_distance_matrix)
from .neighbors import (BalancedKNN, connectivity_to_weights,
                                convolve_by_sparse_weights,
                                knn_distance_matrix)
#from velocyto.serialization import dump_hdf5, load_hdf5


#from tqdm.auto import tqdm
from .oracle_utility import _adata_to_matrix, _adata_to_df, _get_clustercolor_from_anndata


class modified_VelocytoLoom():

    def __init__(self):

        pass



    def score_cv_vs_mean(self, N: int=3000, min_expr_cells: int=2, max_expr_avg: float=20, min_expr_avg: int=0, svr_gamma: float=None,
                         winsorize: bool=False, winsor_perc: Tuple[float, float]=(1, 99.5), sort_inverse: bool=False, plot: bool=False) -> np.ndarray:
        from sklearn.svm import SVR
        """Rank genes on the basis of a CV vs mean fit, it uses a nonparametric fit (Support Vector Regression)

        Arguments
        ---------
        N: int
            the number to select
        min_expr_cells: int, (default=2)
            minimum number of cells that express that gene for it to be considered in the fit
        min_expr_avg: int, (default=0)
            The minimum average accepted before discarding from the the gene as not expressed
        max_expr_avg: float, (default=20)
            The maximum average accepted before discarding from the the gene as house-keeping/outlier
        svr_gamma: float
            the gamma hyper-parameter of the SVR
        winsorize: bool
            Wether to winsorize the data for the cv vs mean model
        winsor_perc: tuple, default=(1, 99.5)
            the up and lower bound of the winsorization
        sort_inverse: bool, (default=False)
            if True it sorts genes from less noisy to more noisy (to use for size estimation not for feature selection)
        which: bool, (default="S")
            it performs the same cv_vs mean procedure on spliced "S" or unspliced "U" count
            "both" is NOT supported here because most often S the two procedure would have different parameters
            (notice that default parameters are good heuristics only for S)
        plot: bool, default=False
            whether to show a plot

        Returns
        -------
        Nothing but it creates the attributes
        cv_mean_score: np.ndarray
            How much the observed CV is higher than the one predicted by a noise model fit to the data
        cv_mean_selected: np.ndarray bool
            on the basis of the N parameter

        Note: genes excluded from the fit will have in the output the same score as the lowest scoring gene in the dataset.

        To perform the filtering use the method `filter_genes`
        """

        X = _adata_to_matrix(self.adata, "raw_count")
        if winsorize:
            if min_expr_cells <= ((100 - winsor_perc[1]) * X.shape[1] * 0.01):
                min_expr_cells = int(np.ceil((100 - winsor_perc[1]) * X.shape[0] * 0.01)) + 2
                logging.debug(f"min_expr_cells is too low for winsorization with upper_perc ={winsor_perc[1]}, upgrading to min_expr_cells ={min_expr_cells}")

        detected_bool = ((X > 0).sum(1) > min_expr_cells) & (X.mean(1) < max_expr_avg) & (X.mean(1) > min_expr_avg)
        Sf = X[detected_bool, :]
        if winsorize:
            down, up = np.percentile(Sf, winsor_perc, 1)
            Sfw = np.clip(Sf, down[:, None], up[:, None])
            mu = Sfw.mean(1)
            sigma = Sfw.std(1, ddof=1)
        else:
            mu = Sf.mean(1)
            sigma = Sf.std(1, ddof=1)

        cv = sigma / mu
        log_m = np.log2(mu)
        log_cv = np.log2(cv)

        if svr_gamma is None:
            svr_gamma = 150. / len(mu)
            logging.debug(f"svr_gamma set to {svr_gamma}")
        # Fit the Support Vector Regression
        clf = SVR(gamma=svr_gamma)
        clf.fit(log_m[:, None], log_cv)
        fitted_fun = clf.predict
        ff = fitted_fun(log_m[:, None])
        score = log_cv - ff
        if sort_inverse:
            score = - score
        nth_score = np.sort(score)[::-1][N]
        if plot:
            scatter_viz(log_m[score > nth_score], log_cv[score > nth_score], s=3, alpha=0.4, c="tab:red")
            scatter_viz(log_m[score <= nth_score], log_cv[score <= nth_score], s=3, alpha=0.4, c="tab:blue")
            mu_linspace = np.linspace(np.min(log_m), np.max(log_m))
            plt.plot(mu_linspace, fitted_fun(mu_linspace[:, None]), c="k")
            plt.xlabel("log2 mean S")
            plt.ylabel("log2 CV S")
        self.cv_mean_score = np.zeros(detected_bool.shape)
        self.cv_mean_score[~detected_bool] = np.min(score) - 1e-16
        self.cv_mean_score[detected_bool] = score
        self.cv_mean_selected = self.cv_mean_score >= nth_score
        self.cv_mean_selected_genes = self.adata.var.index[self.cv_mean_selected].values



    def perform_PCA(self, n_components: int=None, div_by_std: bool=False) -> None:
        """Perform PCA (cells as samples)

        Arguments
        ---------
        which: str, default="S_norm"
            The name of the attribute to use for the calculation (e.g. S_norm or Sx_norm)
        n_components: int, default=None
            Number of components to keep. If None all the components will be kept.
        div_by_std: bool, default=False
            Wether to divide by standard deviation

        Returns
        -------
        Returns nothing but it creates the attributes:
        pca: np.ndarray
            a numpy array of shape (cells, npcs)

        """
        X = _adata_to_matrix(self.adata, "normalized_count")

        self.pca = PCA(n_components=n_components)
        if div_by_std:
            self.pcs = self.pca.fit_transform(X.T / X.std(0))
        else:
            self.pcs = self.pca.fit_transform(X.T)



    def plot_pca(self, dim: List[int]=[0, 1, 2], elev: float=60, azim: float=-140) -> None:
        """Plot 3d PCA
        """

        # update color information
        col_dict = _get_clustercolor_from_anndata(adata=self.adata,
                                                  cluster_name=self.cluster_column_name,
                                                  return_as="dict")
        self.colorandum = np.array([col_dict[i] for i in self.adata.obs[self.cluster_column_name]])

        fig = plt.figure(figsize=(8, 6))
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(self.pcs[:, dim[0]],
                   self.pcs[:, dim[1]],
                   self.pcs[:, dim[2]],
                   c=self.colorandum)
        ax.view_init(elev=elev, azim=azim)



    def knn_imputation(self, k: int=None, metric: str="euclidean", diag: float=1,
                       n_pca_dims: int=None, maximum: bool=False,
                       balanced: bool=False, b_sight: int=None, b_maxl: int=None,
                       group_constraint: Union[str, np.ndarray]=None, n_jobs: int=8) -> None:
        """Performs k-nn smoothing of the data matrix

        Arguments
        ---------
        k: int
            number of neighbors. If None the default it is chosen to be `0.025 * Ncells`
        metric: str
            "euclidean" or "correlation"
        diag: int, default=1
            before smoothing this value is substituted in the diagonal of the knn contiguity matrix
            Resulting in a reduction of the smoothing effect.
            E.g. if diag=8 and k=10 value of Si = (8 * S_i + sum(S_n, with n in 5nn of i)) / (8+5)
        maximum: bool, default=False
            If True the maximum value of the smoothing and the original matrix entry is taken.
        n_pca_dims: int, default=None
            number of pca to use for the knn distance metric. If None all pcs will be used. (used only if pca_space == True)
        balanced: bool
            whether to use BalancedKNN version
        b_sight: int
            the sight parameter of BalancedKNN (used only if balanced == True)
        b_maxl: int
            the maxl parameter of BalancedKNN (used only if balanced == True)

        n_jobs: int, default 8
            number of parallel jobs in knn calculation

        Returns
        -------
        Nothing but it creates the attributes:
        knn: scipy.sparse.csr_matrix
            knn contiguity matrix
        knn_smoothing_w: scipy.sparse.lil_matrix
            the weights used for the smoothing
        Sx: np.ndarray
            smoothed spliced
        Ux: np.ndarray
            smoothed unspliced

        """
        X = _adata_to_matrix(self.adata, "normalized_count")


        N = self.adata.shape[0] # cell number

        if k is None:
            k = int(N * 0.025)
        if b_sight is None and balanced:
            b_sight = int(k * 8)
        if b_maxl is None and balanced:
            b_maxl = int(k * 4)

        space = self.pcs[:, :n_pca_dims]

        if balanced:
            bknn = BalancedKNN(k=k, sight_k=b_sight, maxl=b_maxl,
                               metric=metric, mode="distance", n_jobs=n_jobs)
            bknn.fit(space)
            self.knn = bknn.kneighbors_graph(mode="distance")
        else:

            self.knn = knn_distance_matrix(space, metric=metric, k=k,
                                           mode="distance", n_jobs=n_jobs)
        connectivity = (self.knn > 0).astype(float)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # SparseEfficiencyWarning: Changing the sparsity structure of a csr_matrix is expensive. lil_matrix is more efficient.
            connectivity.setdiag(diag)
        self.knn_smoothing_w = connectivity_to_weights(connectivity)

        ###
        Xx = convolve_by_sparse_weights(X, self.knn_smoothing_w)
        self.adata.layers["imputed_count"] = Xx.transpose().copy()

        self.k_knn_imputation = k


    def estimate_transition_prob(self,
                                 n_neighbors: int=None,
                                 knn_random: bool=True, sampled_fraction: float=0.3,
                                 sampling_probs: Tuple[float, float]=(0.5, 0.1),
                                 n_jobs: int=4, threads: int=None, calculate_randomized: bool=True,
                                 random_seed: int=15071990, cell_idx_use=None) -> None:
        """Use correlation to estimate transition probabilities for every cells to its embedding neighborhood

        Arguments
        ---------
        embed: str, default="ts"
            The name of the attribute containing the embedding. It will be retrieved as getattr(self, embed)
        transform: str, default="sqrt"
            The transformation that is applies on the high dimensional space.
            If None the raw data will be used

        n_sight: int, default=None (also n_neighbors)
            The number of neighbors to take into account when performing the projection
        knn_random: bool, default=True
            whether to random sample the neighborhoods to speedup calculation
        sampling_probs: Tuple, default=(0.5, 1)
        max_dist_embed: float, default=None
            CURRENTLY NOT USED
            The maximum distance allowed
            If None it will be set to 0.25 * average_distance_two_points_taken_at_random
        n_jobs: int, default=4
            number of jobs to calculate knn
            this only applies to the knn search, for the more time consuming correlation computation see threads
        threads: int, default=None
            The threads will be used for the actual correlation computation by default half of the total.
        calculate_randomized: bool, default=True
            Calculate the transition probabilities with randomized residuals.
            This can be plotted downstream as a negative control and can be used to adjust the visualization scale of the velocity field.
        random_seed: int, default=15071990
            Random seed to make knn_random mode reproducible

        Returns
        -------
        """

        numba_random_seed(random_seed)

        X = _adata_to_matrix(self.adata, "imputed_count")  # [:, :ndims]
        delta_X = _adata_to_matrix(self.adata, "delta_X")
        embedding = self.adata.obsm[self.embedding_name]
        self.embedding = embedding

        if n_neighbors is None:
            n_neighbors = int(self.adata.shape[0] / 5)


        if knn_random:
            np.random.seed(random_seed)
            self.corr_calc = "knn_random"


            if calculate_randomized:
                delta_X_rndm = np.copy(delta_X)
                permute_rows_nsign(delta_X_rndm)

            logging.debug("Calculate KNN in the embedding space")

            if cell_idx_use is None:
                nn = NearestNeighbors(n_neighbors=n_neighbors + 1, n_jobs=n_jobs)
                nn.fit(embedding)  # NOTE should support knn in high dimensions
                self.embedding_knn = nn.kneighbors_graph(mode="connectivity")

            else:
                self.embedding_knn = calculate_embedding_knn_with_cell_idx(embedding_original=self.embedding,
                                                                           cell_idx_use=cell_idx_use,
                                                                           n_neighbors=n_neighbors,
                                                                           n_jobs=n_jobs)

            # Pick random neighbours and prune the rest
            neigh_ixs = self.embedding_knn.indices.reshape((-1, n_neighbors + 1))
            p = np.linspace(sampling_probs[0], sampling_probs[1], neigh_ixs.shape[1])
            p = p / p.sum()

            # There was a problem of API consistency because the random.choice can pick the diagonal value (or not)
            # resulting self.corrcoeff with different number of nonzero entry per row.
            # Not updated yet not to break previous analyses
            # Fix is substituting below `neigh_ixs.shape[1]` with `np.arange(1,neigh_ixs.shape[1]-1)`
            # I change it here since I am doing some breaking changes
            sampling_ixs = np.stack([np.random.choice(neigh_ixs.shape[1],
                                                      size=(int(sampled_fraction * (n_neighbors + 1)),),
                                                      replace=False,
                                                      p=p) for i in range(neigh_ixs.shape[0])], 0)
            self.sampling_ixs = sampling_ixs
            neigh_ixs = neigh_ixs[np.arange(neigh_ixs.shape[0])[:, None], sampling_ixs]
            nonzero = neigh_ixs.shape[0] * neigh_ixs.shape[1]
            self.embedding_knn = sparse.csr_matrix((np.ones(nonzero),
                                                    neigh_ixs.ravel(),
                                                    np.arange(0, nonzero + 1, neigh_ixs.shape[1])),
                                                   shape=(neigh_ixs.shape[0],
                                                          neigh_ixs.shape[0]))

            logging.debug(f"Correlation Calculation '{self.corr_calc}'")

            ###
            ###
            self.corrcoef = colDeltaCorpartial(X, delta_X, neigh_ixs, threads=threads)
            if calculate_randomized:
                logging.debug(f"Correlation Calculation for negative control")
                self.corrcoef_random = colDeltaCorpartial(X, delta_X_rndm, neigh_ixs, threads=threads)
            ######

            if np.any(np.isnan(self.corrcoef)):
                self.corrcoef[np.isnan(self.corrcoef)] = 1
                logging.debug("Nans encountered in corrcoef and corrected to 1s. If not identical cells were present it is probably a small isolated cluster converging after imputation.")
                #logging.warning("Nans encountered in corrcoef and corrected to 1s. If not identical cells were present it is probably a small isolated cluster converging after imputation.")
            if calculate_randomized:
                np.fill_diagonal(self.corrcoef_random, 0)
                if np.any(np.isnan(self.corrcoef_random)):
                    self.corrcoef_random[np.isnan(self.corrcoef_random)] = 1
                    #logging.warning("Nans encountered in corrcoef and corrected to 1s. If not identical cells were present it is probably a small isolated cluster converging after imputation.")
                    logging.debug("Nans encountered in corrcoef_random and corrected to 1s. If not identical cells were present it is probably a small isolated cluster converging after imputation.")
            logging.debug(f"Done Correlation Calculation")
        else:
            self.corr_calc = "full"

            if calculate_randomized:
                delta_X_rndm = np.copy(delta_X)
                permute_rows_nsign(delta_X_rndm)

            logging.debug("Calculate KNN in the embedding space")
            nn = NearestNeighbors(n_neighbors=n_neighbors + 1, n_jobs=n_jobs)
            nn.fit(embedding)
            self.embedding_knn = nn.kneighbors_graph(mode="connectivity")

            logging.debug("Correlation Calculation 'full'")
            #####
            self.corrcoef = colDeltaCor(X, delta_X, threads=threads)
            if calculate_randomized:
                logging.debug(f"Correlation Calculation for negative control")
                self.corrcoef_random = colDeltaCor(X, delta_X_rndm, threads=threads)

            #####
            np.fill_diagonal(self.corrcoef, 0)
            if calculate_randomized:
                np.fill_diagonal(self.corrcoef_random, 0)

    def calculate_embedding_shift(self, sigma_corr: float=0.05) -> None:
        """Use the transition probability to project the velocity direction on the embedding

        Arguments
        ---------
        sigma_corr: float, default=0.05
            the kernel scaling

        Returns
        -------
        Nothing but it creates the following attributes:
        transition_prob: np.ndarray
            the transition probability calculated using the exponential kernel on the correlation coefficient
        delta_embedding: np.ndarray
            The resulting vector
        """
        # Kernel evaluation
        logging.debug("Calculate transition probability")

        # NOTE maybe sparse matrix here are slower than dense
        # NOTE if knn_random this could be made much faster either using sparse matrix or neigh_ixs
        self.transition_prob = np.exp(self.corrcoef / sigma_corr) * self.embedding_knn.A  # naive
        self.transition_prob /= self.transition_prob.sum(1)[:, None]
        if hasattr(self, "corrcoef_random"):
            logging.debug("Calculate transition probability for negative control")
            self.transition_prob_random = np.exp(self.corrcoef_random / sigma_corr) * self.embedding_knn.A  # naive
            self.transition_prob_random /= self.transition_prob_random.sum(1)[:, None]

        unitary_vectors = self.embedding.T[:, None, :] - self.embedding.T[:, :, None]  # shape (2,ncells,ncells)
        with np.errstate(divide='ignore', invalid='ignore'):
            unitary_vectors /= np.linalg.norm(unitary_vectors, ord=2, axis=0)  # divide by L2
            np.fill_diagonal(unitary_vectors[0, ...], 0)  # fix nans
            np.fill_diagonal(unitary_vectors[1, ...], 0)

        self.delta_embedding = (self.transition_prob * unitary_vectors).sum(2)
        self.delta_embedding -= (self.embedding_knn.A * unitary_vectors).sum(2) / self.embedding_knn.sum(1).A.T
        self.delta_embedding = self.delta_embedding.T


        if hasattr(self, "corrcoef_random"):
            self.delta_embedding_random = (self.transition_prob_random * unitary_vectors).sum(2)
            self.delta_embedding_random -= (self.embedding_knn.A * unitary_vectors).sum(2) / self.embedding_knn.sum(1).A.T
            self.delta_embedding_random = self.delta_embedding_random.T


    def calculate_grid_arrows(self, smooth: float=0.5, steps: Tuple=(40, 40),
                              n_neighbors: int=100, n_jobs: int=4, xylim: Tuple=((None, None), (None, None))) -> None:
        """Calculate the velocity using a points on a regular grid and a gaussian kernel

        Note: the function should work also for n-dimensional grid

        Arguments
        ---------
        embed: str, default=embedding
            The name of the attribute containing the embedding. It will be retrieved as getattr(self, embed)
            The difference vector is getattr(self, 'delta' + '_' + embed)
        smooth: float, smooth=0.5
            Higher value correspond to taking in consideration further points
            the standard deviation of the gaussian kernel is smooth * stepsize
        steps: tuple, default
            the number of steps in the grid for each axis
        n_neighbors:
            number of neighbors to use in the calculation, bigger number should not change too much the results..
            ...as soon as smooth is small
            Higher value correspond to slower execution time
        n_jobs:
            number of processes for parallel computing
        xymin:
            ((xmin, xmax), (ymin, ymax))

        Returns
        -------
        Nothing but it sets the attributes:
        flow_embedding: np.ndarray
            the coordinates of the embedding
        flow_grid: np.ndarray
            the gridpoints
        flow: np.ndarray
            vector field coordinates
        flow_magnitude: np.ndarray
            magnitude of each vector on the grid
        total_p_mass: np.ndarray
            density at each point of the grid

        """
        embedding = self.embedding
        delta_embedding = getattr(self, f"delta_embedding")

        if hasattr(self, "corrcoef_random"):
            delta_embedding_random = getattr(self, f"delta_embedding_random")

        # Prepare the grid
        grs = []
        for dim_i in range(embedding.shape[1]):
            m, M = np.min(embedding[:, dim_i]), np.max(embedding[:, dim_i])

            if xylim[dim_i][0] is not None:
                m = xylim[dim_i][0]
            if xylim[dim_i][1] is not None:
                M = xylim[dim_i][1]

            m = m - 0.025 * np.abs(M - m)
            M = M + 0.025 * np.abs(M - m)
            gr = np.linspace(m, M, steps[dim_i])
            grs.append(gr)

        meshes_tuple = np.meshgrid(*grs)
        gridpoints_coordinates = np.vstack([i.flat for i in meshes_tuple]).T

        nn = NearestNeighbors(n_neighbors=n_neighbors, n_jobs=n_jobs)
        nn.fit(embedding)
        dists, neighs = nn.kneighbors(gridpoints_coordinates)

        std = np.mean([(g[1] - g[0]) for g in grs])
        # isotropic gaussian kernel
        gaussian_w = normal.pdf(loc=0, scale=smooth * std, x=dists)
        self.total_p_mass = gaussian_w.sum(1)

        UZ = (delta_embedding[neighs] * gaussian_w[:, :, None]).sum(1) / np.maximum(1, self.total_p_mass)[:, None]  # weighed average
        magnitude = np.linalg.norm(UZ, axis=1)
        # Assign attributes
        self.flow_embedding = embedding
        self.flow_grid = gridpoints_coordinates
        self.flow = UZ
        self.flow_norm = UZ / np.percentile(magnitude, 99.5)
        self.flow_norm_magnitude = np.linalg.norm(self.flow_norm, axis=1)

        if hasattr(self, "corrcoef_random"):
            UZ_rndm = (delta_embedding_random[neighs] * gaussian_w[:, :, None]).sum(1) / np.maximum(1, self.total_p_mass)[:, None]  # weighed average
            magnitude_rndm = np.linalg.norm(UZ, axis=1)
            # Assign attributes
            self.flow_rndm = UZ_rndm
            self.flow_norm_rndm = UZ_rndm / np.percentile(magnitude_rndm, 99.5)
            self.flow_norm_magnitude_rndm = np.linalg.norm(self.flow_norm_rndm, axis=1)

    def prepare_markov(self, sigma_D: np.ndarray, sigma_W: np.ndarray, direction: str="forward", cells_ixs: np.ndarray=None) -> None:
        """Prepare a transition probability for Markov process

        Arguments
        ---------
        sigma_D: float
            the standard deviation used on the locality-limiting component
        sigma_W: float
            the standard deviation used on the noise component
        direction: str, default="backwards"
            whether to diffuse forward of backwards
        cells_ixs: np.ndarray, default=None
            Cells to use, if None all the cells will be considered.

        Returns
        -------
        Nothing but it creates the following attributes:
        tr: np.ndarray
            the transition probability matrix

        """
        if cells_ixs is None:
            cells_ixs = np.arange(self.transition_prob.shape[0])

        # NOTE: This implementation is not speed optimized to improve the speed of the implementation:
        # - the C/Fortran contiguity of the transition matrix should be taken into account
        # - a knn implementation would reduce computation
        # - should avoid transformation to and from dense-sparse formats
        if direction == "forward":
            self.tr = np.array(self.transition_prob[cells_ixs, :][:, cells_ixs])
        elif direction == "backwards":
            self.tr = np.array((self.transition_prob[cells_ixs, :][:, cells_ixs]).T, order="C")
        else:
            raise NotImplementedError(f"{direction} is not an implemented direction")
        dist_matrix = squareform(pdist(self.embedding[cells_ixs, :]))
        K_D = gaussian_kernel(dist_matrix, sigma=sigma_D)
        self.tr = self.tr * K_D
        # Fill diagonal with max or the row and sum=1 normalize
        np.fill_diagonal(self.tr, self.tr.max(1))
        self.tr = self.tr / self.tr.sum(1)[:, None]

        K_W = gaussian_kernel(dist_matrix, sigma=sigma_W)
        K_W = K_W / K_W.sum(1)[:, None]
        self.tr = 0.8 * self.tr + 0.2 * K_W
        self.tr = self.tr / self.tr.sum(1)[:, None]
        self.tr = scipy.sparse.csr_matrix(self.tr)

        if hasattr(self, "corrcoef_random"):
            if direction == "forward":
                self.tr_random = np.array(self.transition_prob_random[cells_ixs, :][:, cells_ixs])
            elif direction == "backwards":
                self.tr_random = np.array((self.transition_prob_random[cells_ixs, :][:, cells_ixs]).T, order="C")
            else:
                raise NotImplementedError(f"{direction} is not an implemented direction")
            #dist_matrix = squareform(pdist(self.embedding[cells_ixs, :]))
            #K_D = gaussian_kernel(dist_matrix, sigma=sigma_D)
            self.tr_random = self.tr_random * K_D
            # Fill diagonal with max or the row and sum=1 normalize
            np.fill_diagonal(self.tr_random, self.tr_random.max(1))
            self.tr_random = self.tr_random / self.tr_random.sum(1)[:, None]

            #K_W = gaussian_kernel(dist_matrix, sigma=sigma_W)
            #K_W = K_W / K_W.sum(1)[:, None]
            self.tr_random = 0.8 * self.tr_random + 0.2 * K_W
            self.tr_random = self.tr_random / self.tr_random.sum(1)[:, None]
            self.tr_random = scipy.sparse.csr_matrix(self.tr_random)

    def run_markov(self, starting_p: np.ndarray=None, n_steps: int=2500, mode: str="time_evolution") -> None:
        """Run a Markov process

        Arguments
        ---------
        starting_p: np.ndarray, default=None
            specifies the starting density
            if None is passed an array of 1/self.tr.shape[0] will be created
        n_steps: np.ndarray, default=2500
            Numbers of steps to be performed
        mode: str, default="time_evolution"
            this argument is passed to the Diffusion.diffuse call

        Returns
        -------
        Nothing but it creates the attribute:
        diffused: np.ndarray
            The probability to be found at any of the states
        """
        self.prepare_markov_simulation()

        if starting_p is None:
            starting_p = np.ones(self.tr.shape[0]) / self.tr.shape[0]
        diffusor = Diffusion()
        self.diffused = diffusor.diffuse(starting_p, self.tr, n_steps=n_steps, mode=mode)[0]

    def plot_mc_results_as_density(self, args={}):
        """
        this function plot the results of mc chain simulation in velocyto method.
        """
        diffused_n = self.diffused - np.percentile(self.diffused, 3)
        diffused_n /= np.percentile(diffused_n, 97)
        diffused_n = np.clip(diffused_n, 0, 1)
        #diffused_n = tv.diffused
        #plt.figure(None,(7,7))
        args_ = {"alpha": 0.5, "s": 50, "lw": 0., "edgecolor":"", "cmap": "viridis_r", "rasterized": True}
        args_.update(args)

        plt.scatter(self.embedding[self.ixs_mcmc, 0], self.embedding[self.ixs_mcmc, 1],
                        c=diffused_n, **args_)
        plt.axis("off")
        cax = plt.axes([0.85, 0.1, 0.075, 0.8])
        plt.colorbar()
        plt.axis("off")



    def plot_grid_arrows(self, quiver_scale: Union[str, float]="auto", scale_type: str= "relative", min_mass: float=1, min_magnitude: float=None,
                         scatter_kwargs_dict: Dict= None, plot_dots: bool=False, plot_random: bool=False, **quiver_kwargs: Any) -> None:
        print("This function is deprecated after celloracle ver 0.9.4. Please see our newest tutorial for the alternative functions. \nhttps://morris-lab.github.io/CellOracle.documentation/tutorials/index.html")

        '''
        """Plots vector field averaging velocity vectors on a grid

        Arguments
        ---------
        quiver_scale: float, default="auto"
            Rescaling factor applied to the arrow field to enhance visibility
            If "auto" the scale is selected using the randomized (negative) control (even if `plot_random`=False)
            If a float is provided the interpretation of the value depends on the parameter `scale_type`, see below.
            NOTE: In the situation where "auto" is resulting in very small or big velocities, pass a float to this parameter
            The float will be interpreted as a scaling, importantly both the data and the control will be scaled
            in this way you can rescale the velocity arbitrarily without the risk of observing just an overfit of the noise
        scale_type: str, default="relative"
            How to interpret `quiver_scale`:
            If "relative" (default) the value will be used as a scaling factor and multiplied by the value from "auto"
            (it follows that quiver_scale="auto" is equivalent to quiver_scale=1)
            If "absolute" the value will be passed to the matplotlib quiver function (not recommended if you are not sure what this implies)
        min_mass: float, default=1
            the minimum density around a grid point for it to be considered and plotted
        min_magnitude: float, default=None
            the minimum magnitude of the velocity for it to be considered and plotted
        scatter_kwargs_dict: dict, default=None
            a dictionary of keyword arguments to pass to scatter
            by default the following are passed: s=20, zorder=-1, alpha=0.2, lw=0, c=self.colorandum. But they can be overridden.
        plot_dots: bool, default=True
            whether to plot dots in correspondence of all low velocity grid points
        plot_random: bool, default=True
            whether to plot the randomized control next to the plot
        **quiver_kwargs: dict
            keyword arguments to pass to quiver
            By default the following are passed angles='xy', scale_units='xy', minlength=1.5. But they can be overridden.
        """
        # update color information
        col_dict = _get_clustercolor_from_anndata(adata=self.adata,
                                                  cluster_name=self.cluster_column_name,
                                                  return_as="dict")
        self.colorandum = np.array([col_dict[i] for i in self.adata.obs[self.cluster_column_name]])

        # plt.figure(figsize=(10, 10))
        _quiver_kwargs = {"angles": 'xy', "scale_units": 'xy', "minlength": 1.5}
        _quiver_kwargs.update(quiver_kwargs)
        scatter_dict = {"s": 20, "zorder": -1, "alpha": 0.2, "lw": 0, "c": self.colorandum}
        if scatter_kwargs_dict is not None:
            scatter_dict.update(scatter_kwargs_dict)

        # Determine quiver scale
        if scale_type == "relative":
            if hasattr(self, "flow_rndm"):
                plot_scale = np.linalg.norm(np.max(self.flow_grid, 0) - np.min(self.flow_grid, 0), 2)  # Diagonal of the plot
                arrows_scale = np.percentile(np.linalg.norm(self.flow_rndm[self.total_p_mass >= min_mass, :], 2, 1), 90)  # Tipical lenght of an arrow
                if quiver_scale == "auto":
                    quiver_scale = arrows_scale / (plot_scale * 0.0025)
                else:
                    quiver_scale = quiver_scale * arrows_scale / (plot_scale * 0.0025)
            else:
                raise ValueError(""""`scale_type` was set to 'relative' but the randomized control was not computed when running estimate_transition_prob
                Please run estimate_transition_prob or set `scale_type` to `absolute`""")
        else:
            logging.warning("The arrow scale was set to be 'absolute' make sure you know how to properly interpret the plots")

        mass_filter = self.total_p_mass < min_mass
        if min_magnitude is None:
            XY, UV = np.copy(self.flow_grid), np.copy(self.flow)
            if not plot_dots:
                UV = UV[~mass_filter, :]
                XY = XY[~mass_filter, :]
            else:
                UV[mass_filter, :] = 0
        else:
            XY, UV = np.copy(self.flow_grid), np.copy(self.flow_norm)
            if not plot_dots:
                UV = UV[~(mass_filter | (self.flow_norm_magnitude < min_magnitude)), :]
                XY = XY[~(mass_filter | (self.flow_norm_magnitude < min_magnitude)), :]
            else:
                UV[mass_filter | (self.flow_norm_magnitude < min_magnitude), :] = 0


        if min_magnitude is None:
            XY, UV_rndm = np.copy(self.flow_grid), np.copy(self.flow_rndm)
            if not plot_dots:
                UV_rndm = UV_rndm[~mass_filter, :]
                XY = XY[~mass_filter, :]
            else:
                UV_rndm[mass_filter, :] = 0
        else:
            XY, UV_rndm = np.copy(self.flow_grid), np.copy(self.flow_norm_rndm)
            if not plot_dots:
                UV_rndm = UV_rndm[~(mass_filter | (self.flow_norm_magnitude_rndm < min_magnitude)), :]
                XY = XY[~(mass_filter | (self.flow_norm_magnitude_rndm < min_magnitude)), :]
            else:
                UV_rndm[mass_filter | (self.flow_norm_magnitude_rndm < min_magnitude), :] = 0

        if plot_random:
            plt.subplot(122)
            plt.title("Randomized")
            plt.scatter(self.flow_embedding[:, 0], self.flow_embedding[:, 1], **scatter_dict)
            plt.quiver(XY[:, 0], XY[:, 1], UV_rndm[:, 0], UV_rndm[:, 1],
                       scale=quiver_scale, zorder=20000, **_quiver_kwargs)
            plt.axis("off")
            plt.subplot(121)
            plt.title("Data")

        plt.scatter(self.flow_embedding[:, 0], self.flow_embedding[:, 1], **scatter_dict)
        plt.quiver(XY[:, 0], XY[:, 1], UV[:, 0], UV[:, 1],
                   scale=quiver_scale, zorder=20000, **_quiver_kwargs)
        plt.axis("off")
        '''

    def plot_arrows_embedding(self, choice: Union[str, int]="auto", quiver_scale: Union[str, float]="auto", scale_type: str="relative",
                              plot_scatter: bool=False, scatter_kwargs: Dict={}, color_arrow: str="cluster",
                              new_fig: bool=False, plot_random: bool=True, **quiver_kwargs: Any) -> None:
        print("This function is deprecated after celloracle ver 0.9.4. Please see our newest tutorial for the alternative functions. \nhttps://morris-lab.github.io/CellOracle.documentation/tutorials/index.html")

        '''
        """Plots velocity on the embedding cell-wise

        Arguments
        ---------
        choice: int, default = "auto"
            the number of cells to randomly pick to plot the arrows (To avoid overcrowding)
        quiver_scale: float, default="auto"
            Rescaling factor applied to the arrow field to enhance visibility
            If "auto" the scale is selected using the randomized (negative) control (even if `plot_random`=False)
            If a float is provided the interpretation of the value depends on the parameter `scale_type`, see below.
            NOTE: Despite a similar option than plot_grid_arrows, here there is no strong motivation to calculate the scale relative to the randomized control
            This is because the randomized doesn't have to have smaller velocity cell-wise, there might be for example
            scattered cells that will have strong velocity but they will, correctly just average out when calculating the average velocity field.
        scale_type: str, default="relative"
            How to interpret `quiver_scale`:
            If "relative" (default) the value will be used as a scaling factor and multiplied by the value from "auto"
            (it follows that quiver_scale="auto" is equivalent to quiver_scale=1)
            If "absolute" the value will be passed to the matplotlib quiver function
        plot_scatter: bool, default = False
            whether to plot the points
        scatter_kwargs: Dict
            A dictionary containing all the keywords arguments to pass to matplotlib scatter
            by default the following are passed: c="0.8", alpha=0.4, s=10, edgecolor=(0, 0, 0, 1), lw=0.3. But they can be overridden.
        color_arrow: str, default = "cluster"
            the color of the arrows, if "cluster" the arrows are colored the same as the cluster
        new_fig: bool, default=False
            whether to create a new figure
        plot_random: bool, default=True
            whether to plot the randomized control next to the plot
        **quiver_kwargs: dict
            keyword arguments to pass to quiver
            By default the following are passed angles='xy', scale_units='xy', minlength=1.5. But they can be overridden.

        Returns
        -------
        Nothing, just plots the tsne with arrows
        """
        # update color information
        col_dict = _get_clustercolor_from_anndata(adata=self.adata,
                                                  cluster_name=self.cluster_column_name,
                                                  return_as="dict")
        self.colorandum = np.array([col_dict[i] for i in self.adata.obs[self.cluster_column_name]])

        if choice == "auto":
            choice = int(self.S.shape[1] / 3)
            logging.warning(f"Only {choice} arrows will be shown to avoid overcrowding, you can choose the exact number setting the `choice` argument")
        _quiver_kwargs = {"angles": 'xy', "scale_units": 'xy', "minlength": 1.5}
        _scatter_kwargs = dict(c="0.8", alpha=0.4, s=10, edgecolor=(0, 0, 0, 1), lw=0.3)
        _scatter_kwargs.update(scatter_kwargs)
        if new_fig:
            if plot_random and hasattr(self, "delta_embedding_random"):
                plt.figure(figsize=(22, 12))
            else:
                plt.figure(figsize=(14, 14))

        ix_choice = np.random.choice(self.embedding.shape[0], size=choice, replace=False)

        # Determine quiver scale
        if scale_type == "relative":
            if hasattr(self, "delta_embedding_random"):
                plot_scale = np.linalg.norm(np.max(self.flow_grid, 0) - np.min(self.flow_grid, 0), 2)  # Diagonal of the plot
                arrows_scale = np.percentile(np.linalg.norm(self.delta_embedding_random, 2, 1), 80)  # Tipical length of an arrow
                if quiver_scale == "auto":
                    quiver_scale = arrows_scale / (plot_scale * 0.005)
                else:
                    quiver_scale = quiver_scale * arrows_scale / (plot_scale * 0.005)
            else:
                raise ValueError("""`scale_type` was set to 'relative' but the randomized control was not computed when running estimate_transition_prob
                Please run estimate_transition_prob or set `scale_type` to `absolute`""")
        else:
            logging.warning("The arrow scale was set to be 'absolute' make sure you know how to properly interpret the plots")

        if color_arrow == "cluster":
            colorandum = self.colorandum[ix_choice, :]
        else:
            colorandum = color_arrow

        _quiver_kwargs.update({"color": colorandum})
        _quiver_kwargs.update(quiver_kwargs)

        if plot_random and hasattr(self, "delta_embedding_random"):
            plt.subplot(122)
            plt.title("Randomized")
            if plot_scatter:
                plt.scatter(self.embedding[:, 0], self.embedding[:, 1], **_scatter_kwargs)
            plt.quiver(self.embedding[ix_choice, 0], self.embedding[ix_choice, 1],
                       self.delta_embedding_random[ix_choice, 0], self.delta_embedding_random[ix_choice, 1],
                       scale=quiver_scale, **_quiver_kwargs)
            plt.axis("off")
            plt.subplot(121)
            plt.title("Data")

        if plot_scatter:
            plt.scatter(self.embedding[:, 0], self.embedding[:, 1], **_scatter_kwargs)

        plt.quiver(self.embedding[ix_choice, 0], self.embedding[ix_choice, 1],
                   self.delta_embedding[ix_choice, 0], self.delta_embedding[ix_choice, 1],
                   scale=quiver_scale, **_quiver_kwargs)
        plt.axis("off")
        '''

    def plot_cell_transitions(self, cell_ix: int=0, alpha: float=0.1, alpha_neigh: float=0.2,
                              cmap_name: str="RdBu_r", plot_arrow: bool=True,
                              mark_cell: bool=True, head_width: int=3) -> None:

        print("This function is deprecated after celloracle ver 0.9.4. Please see our newest tutorial for the alternative functions. \nhttps://morris-lab.github.io/CellOracle.documentation/tutorials/index.html")

        '''

        """Plot the probability of a cell to transition to any other cell

        This function is untested
        """
        cmap = plt.cm.get_cmap(name=cmap_name)
        colorandum = np.ones((self.embedding.shape[0], 4))
        colorandum *= 0.3
        colorandum[:, -1] = alpha

        plt.scatter(self.embedding[:, 0], self.embedding[:, 1],
                    c=colorandum, s=50, edgecolor="")
        if mark_cell:
            plt.scatter(self.embedding[cell_ix, 0], self.embedding[cell_ix, 1],
                        facecolor="none", s=100, edgecolor="k")
        if plot_arrow:
            plt.arrow(self.embedding[cell_ix, 0], self.embedding[cell_ix, 1],
                      self.delta_embedding[cell_ix, 0], self.delta_embedding[cell_ix, 1],
                      head_width=head_width, length_includes_head=True)
        '''




def scatter_viz(x: np.ndarray, y: np.ndarray, *args: Any, **kwargs: Any) -> Any:
    """A wrapper of scatter plot that guarantees that every point is visible in a very crowded scatterplot

    Args
    ----
    x: np.ndarray
        x axis coordinates
    y: np.ndarray
        y axis coordinates
    args and kwargs:
        positional and keyword arguments as in matplotplib.pyplot.scatter

    Returns
    -------
    Plots the graph and returns the axes object
    """
    ix_x_sort = np.argsort(x, kind="mergesort")
    ix_yx_sort = np.argsort(y[ix_x_sort], kind="mergesort")
    args_new = []
    kwargs_new = {}
    for arg in args:
        if type(arg) is np.ndarray:
            args_new.append(arg[ix_x_sort][ix_yx_sort])
        else:
            args_new.append(arg)
    for karg, varg in kwargs.items():
        if type(varg) is np.ndarray:
            kwargs_new[karg] = varg[ix_x_sort][ix_yx_sort]
        else:
            kwargs_new[karg] = varg
    ax = plt.scatter(x[ix_x_sort][ix_yx_sort], y[ix_x_sort][ix_yx_sort], *args_new, **kwargs_new)
    return ax



def calculate_embedding_knn_with_cell_idx(embedding_original, cell_idx_use, n_neighbors, n_jobs=4):

    """
    Calculate knn graph focusing on a cell population.

    """


    nn = NearestNeighbors(n_neighbors=n_neighbors + 1, n_jobs=n_jobs)
    nn.fit(embedding_original[cell_idx_use, :])  # NOTE should support knn in high dimensions
    embedding_knn = nn.kneighbors_graph(mode="connectivity")

    #print(embedding_knn.indices.max())

    indices_in_original_emb = cell_idx_use[embedding_knn.indices]
    neigh_ixs = np.zeros((embedding_original.shape[0], n_neighbors + 1))
    neigh_ixs[cell_idx_use, :] = indices_in_original_emb.reshape((-1, n_neighbors + 1))

    nonzero = neigh_ixs.shape[0] * neigh_ixs.shape[1]
    embedding_knn = sparse.csr_matrix((np.ones(nonzero),
                                      neigh_ixs.ravel(),
                                      np.arange(0, nonzero + 1, neigh_ixs.shape[1])),
                                      shape=(neigh_ixs.shape[0],
                                             neigh_ixs.shape[0]))
    return embedding_knn


@jit(nopython=True)
def numba_random_seed(value: int) -> None:
    """Same as np.random.seed but for numba"""
    np.random.seed(value)


@jit(nopython=True)
def permute_rows_nsign(A: np.ndarray) -> None:
    """Permute in place the entries and randomly switch the sign for each row of a matrix independently.
    """
    plmi = np.array([+1, -1])
    for i in range(A.shape[0]):
        np.random.shuffle(A[i, :])
        A[i, :] = A[i, :] * np.random.choice(plmi, size=A.shape[1])



def gaussian_kernel(X: np.ndarray, mu: float=0, sigma: float=1) -> np.ndarray:
    """Compute gaussian kernel"""
    return np.exp(-(X - mu)**2 / (2 * sigma**2)) / np.sqrt(2 * np.pi * sigma**2)



 # -*- coding: utf-8 -*-
'''


'''
import pandas as pd
import numpy as np

from urllib import request
import sys
import os


from goatools.go_enrichment import GOEnrichmentStudy
from goatools.obo_parser import GODag
from goatools.associations import read_ncbi_gene2go

go_folder = os.path.join(os.path.dirname(__file__), "data")
Xtable_human=pd.read_csv(os.path.join(go_folder,'hg19_xref.txt'), sep='\t')
Xtable_mouse=pd.read_csv(os.path.join(go_folder,'biomart_xref.mm10.txt'), sep='\t')

def _check_data_and_download_if_necessary(data_folder):
    files = os.listdir(data_folder)

    if not 'gene2go.txt' in files:
        print("gene2go file was not found in the PC. Downloading gene2go file from ncbi ...")
        path = os.path.join(data_folder, "gene2go.txt.gz")
        #os.system(f"wget -O {path} https://ftp.ncbi.nlm.nih.gov/gene/DATA/gene2go.gz")
        request.urlretrieve("https://ftp.ncbi.nlm.nih.gov/gene/DATA/gene2go.gz", path)
        os.system(f"gunzip {path}")

    if not 'go-basic.obo' in files:
        print("go-basic file was not found in the PC. Downloading gene2go file from geneontology.org ...")
        path = os.path.join(data_folder, "go-basic.obo")
        os.system(f"wget -O {path} http://geneontology.org/ontology/go-basic.obo")


def geneSymbol2ID(symbols, species="mouse"):
    """
    Convert gene symbol into Entrez gene id.

    Args:
        symbols (array of str): gene symbol

        species (str): Select species. Either "mouse" or "human"

    Returns:
        list of str: Entrez gene id

    """

    if (species=='human'):
        Xtable=pd.read_csv(os.path.join(go_folder,'hg19_xref.txt'), sep='\t')

    elif(species=='mouse'):
        Xtable=pd.read_csv(os.path.join(go_folder,'biomart_xref.mm10.txt'), sep='\t')

    Xtable=Xtable[['Associated Gene Name','EntrezGene ID']].dropna()
    Xtable.index=Xtable['Associated Gene Name']
    GOIs_entrez=[x for x in np.unique(Xtable.loc[symbols].dropna()['EntrezGene ID'])]

    return GOIs_entrez


def geneID2Symbol(IDs, species="mouse"):
    """
    Convert Entrez gene id into gene symbol.

    Args:
        IDs (array of str): Entrez gene id.

        species (str): Select species. Either "mouse" or "human".

    Returns:
        list of str: Gene symbol

    """

    if (species=='human'):
        Xtable=Xtable_human

    elif(species=='mouse'):
        Xtable=Xtable_mouse

    Xtable=Xtable[['Associated Gene Name','EntrezGene ID']].dropna()
    Xtable.index=Xtable['EntrezGene ID']
    symbols=[x for x in np.unique(Xtable.loc[IDs].dropna()['Associated Gene Name'])]

    return symbols

def _ids2symbols(study_ids, species):

    if study_ids == "":
        return []
    ids = study_ids.replace(" ", "").split(",")
    ids = [int(i) for i in ids]
    genes = geneID2Symbol(ids, species)

    return genes


def get_GO(gene_query, species='mouse'):
    """
    Get Gene Ontologies (GOs).

    Args:
        gene_query (array of str): gene list.

        species (str): Select species. Either "mouse" or "human"

    Returns:
        pandas.dataframe: GO analysis results as dataframe.
    """

    sig_thresh = 3
    num_genes = None
    GOIs = gene_query

    # prepare files
    # check files
    _check_data_and_download_if_necessary(go_folder)


    obodag = GODag(os.path.join(go_folder, "go-basic.obo"))

    #go analysis

    if (species=='human'):

        geneid2gos = read_ncbi_gene2go(os.path.join(go_folder,"gene2go.txt"), taxids=[9606])
        print("{N:,} annotated genes".format(N=len(geneid2gos)))

        Xtable=pd.read_csv(os.path.join(go_folder,'hg19_xref.txt'), sep='\t')
        Xtable.index=Xtable['Approved Symbol']
        GOIs_entrez=[int(x) for x in np.unique(Xtable.loc[GOIs].dropna()['EntrezGene ID'])]

    elif (species=='mouse'):

        geneid2gos = read_ncbi_gene2go(os.path.join(go_folder,"gene2go.txt"), taxids=[10090])
        print("{N:,} annotated genes".format(N=len(geneid2gos)))

        from goatools.test_data.genes_NCBI_10090_ProteinCoding import GENEID2NT as GeneID2nt_mus

        Xtable=pd.read_csv(os.path.join(go_folder,'biomart_xref.mm10.txt'), sep='\t')
        Xtable=Xtable[['Associated Gene Name','EntrezGene ID']].dropna()
        Xtable.index=Xtable['Associated Gene Name']
        GOIs_entrez=[int(x) for x in np.unique(Xtable.loc[GOIs].dropna()['EntrezGene ID'])]


    print("processing " + str(len(GOIs)) + " genes ...")


    goeaobj = GOEnrichmentStudy(
        GeneID2nt_mus.keys(), # List of mouse protein-coding genes
        geneid2gos, # geneid/GO associations
        obodag, # Ontologies
        propagate_counts = False,
        alpha = 0.05, # default significance cut-off
        methods = ['fdr_bh']) # defult multipletest correction method


    goea_results = goeaobj.run_study(GOIs_entrez)

    li=[]
    names=[]

    go_default_output = goea_results[0].get_prtflds_default()


    for i in goea_results:
        li.append(i.get_field_values(go_default_output))
        names.append(i.name)

    df_GO = pd.DataFrame(li)

    if len(li) != 0:
        df_GO.columns = go_default_output
        df_GO["genes"] = df_GO.study_items.apply(lambda x: _ids2symbols(x, species))
    else:
        print("Found No GO with significant p-value")

    return df_GO



# -*- coding: utf-8 -*-
'''

'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import pandas as pd
import numpy as np

import sys, os

from tqdm.auto import tqdm
from glob import glob

# 0.2. libraries for DNA and genome data wrangling and Motif analysis

from gimmemotifs.motif import read_motifs

from ..data import __path__ as parent_path
from ..data.config import CELLORACLE_DATA_DIR, WEB_PAR_DIR
from ..utility.data_download_from_web import download_data_if_data_not_exist



MOTIFS_LIST = ['CisBP_ver2_Arabidopsis_thaliana.pfm',
               'CisBP_ver2_Arabidopsis_thaliana_GENE_ID.pfm',
               'CisBP_ver2_Caenorhabditis_elegans.pfm',
               'CisBP_ver2_Cavia_porcellus.pfm',
               'CisBP_ver2_Danio_rerio.pfm',
               'CisBP_ver2_Drosophila_ananassae.pfm',
               'CisBP_ver2_Drosophila_erecta.pfm',
               'CisBP_ver2_Drosophila_grimshawi.pfm',
               'CisBP_ver2_Drosophila_melanogaster.pfm',
               'CisBP_ver2_Drosophila_mix.pfm',
               'CisBP_ver2_Drosophila_mojavensis.pfm',
               'CisBP_ver2_Drosophila_persimilis.pfm',
               'CisBP_ver2_Drosophila_pseudoobscura.pfm',
               'CisBP_ver2_Drosophila_sechellia.pfm',
               'CisBP_ver2_Drosophila_simulans.pfm',
               'CisBP_ver2_Drosophila_virilis.pfm',
               'CisBP_ver2_Drosophila_willistoni.pfm',
               'CisBP_ver2_Drosophila_yakuba.pfm',
               'CisBP_ver2_Gallus_gallus.pfm',
               'CisBP_ver2_Homo_sapiens.pfm',
               'CisBP_ver2_Mus_musculus.pfm',
               'CisBP_ver2_Rattus_norvegicus.pfm',
               'CisBP_ver2_Saccharomyces_cerevisiae.pfm',
               'CisBP_ver2_Sus_scrofa.pfm',
               'CisBP_ver2_Xenopus_laevis.pfm',
               'CisBP_ver2_Xenopus_tropicalis.pfm',
               'CisBP_ver2_Xenopus_tropicalis_and_Xenopus_laevis.pfm']

def load_motifs(motifs_name, force_download=False):
    """

    Load motifs from celloracle motif database

    Args:
        motifs_name (str) : Name of motifs.

    Returns:
        list : List of gimmemotifs.motif object.


    """

    if motifs_name not in MOTIFS_LIST:
        raise ValueError("The motifs name was not in the list. Available motifs: ", MOTIFS_LIST)

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], "motif_data", motifs_name)
    path_factor_data = os.path.join(parent_path[0], "motif_data", motifs_name.replace(".pfm", ".motif2factors.txt"))
    if (force_download == False) & os.path.isfile(path) & os.path.isfile(path_factor_data):
        pass
    else:
        # Download pfm file
        path = os.path.join(CELLORACLE_DATA_DIR, "motif_data", motifs_name)
        backup_url = os.path.join(WEB_PAR_DIR, "motif_data", motifs_name)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)
        # Download motif2factors.txt file
        path_factor_data = os.path.join(CELLORACLE_DATA_DIR, "motif_data", motifs_name.replace(".pfm", ".motif2factors.txt"))
        backup_url = os.path.join(WEB_PAR_DIR, "motif_data", motifs_name.replace(".pfm", ".motif2factors.txt"))
        download_data_if_data_not_exist(path=path_factor_data, backup_url=backup_url)

    motifs = read_motifs(path)

    return motifs


def _make_pwm_file_list(PATH_TO_MOTIF_DATA=None):
    if PATH_TO_MOTIF_DATA is None:
        from ..data import __path__ as data_folder_path
        PATH_TO_MOTIF_DATA = os.path.join(list(data_folder_path)[0], "motif_data")

    PFM_PATHS = glob(os.path.join(PATH_TO_MOTIF_DATA, "*.pfm"))
    PFM_PATHS.sort()
    MOTIFS_PATH_DICT = {os.path.basename(path):path for path in PFM_PATHS}
    MOTIFS_LIST = list(MOTIFS_PATH_DICT.keys())

    return MOTIFS_LIST

####
###



# -*- coding: utf-8 -*-
import pandas as pd

import celloracle as co

# CisBP_ver2_Arabidopsis_thaliana
def test_load_motifs_CisBP_ver2_Arabidopsis_thaliana():
    motifs_name = 'CisBP_ver2_Arabidopsis_thaliana.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Arabidopsis_thaliana():
    motifs_name = 'CisBP_ver2_Arabidopsis_thaliana.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Arabidopsis_thaliana_GENE_ID
def test_load_motifs_CisBP_ver2_Arabidopsis_thaliana_GENE_ID():
    motifs_name = 'CisBP_ver2_Arabidopsis_thaliana_GENE_ID.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Arabidopsis_thaliana_GENE_ID():
    motifs_name = 'CisBP_ver2_Arabidopsis_thaliana_GENE_ID.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Caenorhabditis_elegans
def test_load_motifs_CisBP_ver2_Caenorhabditis_elegans():
    motifs_name = 'CisBP_ver2_Caenorhabditis_elegans.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Caenorhabditis_elegans():
    motifs_name = 'CisBP_ver2_Caenorhabditis_elegans.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Cavia_porcellus
def test_load_motifs_CisBP_ver2_Cavia_porcellus():
    motifs_name = 'CisBP_ver2_Cavia_porcellus.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Cavia_porcellus():
    motifs_name = 'CisBP_ver2_Cavia_porcellus.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Danio_rerio
def test_load_motifs_CisBP_ver2_Danio_rerio():
    motifs_name = 'CisBP_ver2_Danio_rerio.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Danio_rerio():
    motifs_name = 'CisBP_ver2_Danio_rerio.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_ananassae
def test_load_motifs_CisBP_ver2_Drosophila_ananassae():
    motifs_name = 'CisBP_ver2_Drosophila_ananassae.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_ananassae():
    motifs_name = 'CisBP_ver2_Drosophila_ananassae.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_erecta
def test_load_motifs_CisBP_ver2_Drosophila_erecta():
    motifs_name = 'CisBP_ver2_Drosophila_erecta.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_erecta():
    motifs_name = 'CisBP_ver2_Drosophila_erecta.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_grimshawi
def test_load_motifs_CisBP_ver2_Drosophila_grimshawi():
    motifs_name = 'CisBP_ver2_Drosophila_grimshawi.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_grimshawi():
    motifs_name = 'CisBP_ver2_Drosophila_grimshawi.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_melanogaster
def test_load_motifs_CisBP_ver2_Drosophila_melanogaster():
    motifs_name = 'CisBP_ver2_Drosophila_melanogaster.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_melanogaster():
    motifs_name = 'CisBP_ver2_Drosophila_melanogaster.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_mix
def test_load_motifs_CisBP_ver2_Drosophila_mix():
    motifs_name = 'CisBP_ver2_Drosophila_mix.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_mix():
    motifs_name = 'CisBP_ver2_Drosophila_mix.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_mojavensis
def test_load_motifs_CisBP_ver2_Drosophila_mojavensis():
    motifs_name = 'CisBP_ver2_Drosophila_mojavensis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_mojavensis():
    motifs_name = 'CisBP_ver2_Drosophila_mojavensis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_persimilis
def test_load_motifs_CisBP_ver2_Drosophila_persimilis():
    motifs_name = 'CisBP_ver2_Drosophila_persimilis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_persimilis():
    motifs_name = 'CisBP_ver2_Drosophila_persimilis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_pseudoobscura
def test_load_motifs_CisBP_ver2_Drosophila_pseudoobscura():
    motifs_name = 'CisBP_ver2_Drosophila_pseudoobscura.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_pseudoobscura():
    motifs_name = 'CisBP_ver2_Drosophila_pseudoobscura.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_sechellia
def test_load_motifs_CisBP_ver2_Drosophila_sechellia():
    motifs_name = 'CisBP_ver2_Drosophila_sechellia.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_sechellia():
    motifs_name = 'CisBP_ver2_Drosophila_sechellia.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_simulans
def test_load_motifs_CisBP_ver2_Drosophila_simulans():
    motifs_name = 'CisBP_ver2_Drosophila_simulans.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_simulans():
    motifs_name = 'CisBP_ver2_Drosophila_simulans.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_virilis
def test_load_motifs_CisBP_ver2_Drosophila_virilis():
    motifs_name = 'CisBP_ver2_Drosophila_virilis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_virilis():
    motifs_name = 'CisBP_ver2_Drosophila_virilis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_willistoni
def test_load_motifs_CisBP_ver2_Drosophila_willistoni():
    motifs_name = 'CisBP_ver2_Drosophila_willistoni.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_willistoni():
    motifs_name = 'CisBP_ver2_Drosophila_willistoni.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Drosophila_yakuba
def test_load_motifs_CisBP_ver2_Drosophila_yakuba():
    motifs_name = 'CisBP_ver2_Drosophila_yakuba.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Drosophila_yakuba():
    motifs_name = 'CisBP_ver2_Drosophila_yakuba.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Gallus_gallus
def test_load_motifs_CisBP_ver2_Gallus_gallus():
    motifs_name = 'CisBP_ver2_Gallus_gallus.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Gallus_gallus():
    motifs_name = 'CisBP_ver2_Gallus_gallus.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Homo_sapiens
def test_load_motifs_CisBP_ver2_Homo_sapiens():
    motifs_name = 'CisBP_ver2_Homo_sapiens.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Homo_sapiens():
    motifs_name = 'CisBP_ver2_Homo_sapiens.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Mus_musculus
def test_load_motifs_CisBP_ver2_Mus_musculus():
    motifs_name = 'CisBP_ver2_Mus_musculus.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Mus_musculus():
    motifs_name = 'CisBP_ver2_Mus_musculus.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Rattus_norvegicus
def test_load_motifs_CisBP_ver2_Rattus_norvegicus():
    motifs_name = 'CisBP_ver2_Rattus_norvegicus.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Rattus_norvegicus():
    motifs_name = 'CisBP_ver2_Rattus_norvegicus.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Saccharomyces_cerevisiae
def test_load_motifs_CisBP_ver2_Saccharomyces_cerevisiae():
    motifs_name = 'CisBP_ver2_Saccharomyces_cerevisiae.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Saccharomyces_cerevisiae():
    motifs_name = 'CisBP_ver2_Saccharomyces_cerevisiae.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Xenopus_laevis
def test_load_motifs_CisBP_ver2_Xenopus_laevis():
    motifs_name = 'CisBP_ver2_Xenopus_laevis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Xenopus_laevis():
    motifs_name = 'CisBP_ver2_Xenopus_laevis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Xenopus_tropicalis
def test_load_motifs_CisBP_ver2_Xenopus_tropicalis():
    motifs_name = 'CisBP_ver2_Xenopus_tropicalis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Xenopus_tropicalis():
    motifs_name = 'CisBP_ver2_Xenopus_tropicalis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)

# CisBP_ver2_Xenopus_tropicalis_and_Xenopus_laevis
def test_load_motifs_CisBP_ver2_Xenopus_tropicalis_and_Xenopus_laevis():
    motifs_name = 'CisBP_ver2_Xenopus_tropicalis_and_Xenopus_laevis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=False)
    assert isinstance(motifs, list)

def test_load_motifs_dl_CisBP_ver2_Xenopus_tropicalis_and_Xenopus_laevis():
    motifs_name = 'CisBP_ver2_Xenopus_tropicalis_and_Xenopus_laevis.pfm'
    motifs = co.motif_analysis.motif_data.load_motifs(motifs_name=motifs_name,
                                                      force_download=True)
    assert isinstance(motifs, list)



# -*- coding: utf-8 -*-


from IPython.core.display import display, Javascript


def init_datatable_mode(pd):
    """Initialize DataTable mode for pandas DataFrame represenation."""

    # configure path to the datatables library using requireJS
    # that way the library will become globally available
    display(Javascript("""
        require.config({
            paths: {
                DT: '//cdn.datatables.net/1.10.19/js/jquery.dataTables.min',
            }
        });
        $('head').append('<link rel="stylesheet" type="text/css" href="//cdn.datatables.net/1.10.19/css/jquery.dataTables.min.css">');
    """))

    def _repr_datatable_(self):
        """Return DataTable representation of pandas DataFrame."""
        # classes for dataframe table (optional)
        classes = ['table', 'table-striped', 'table-bordered']

        # create table DOM
        script = (
            f'$(element).html(`{self.to_html(index=False, classes=classes)}`);\n'
        )

        # execute jQuery to turn table into DataTable
        script += """
            require(["DT"], function(DT) {
                $(document).ready( () => {
                    // Turn existing table into datatable
                    $(element).find("table.dataframe").DataTable();
                })
            });
        """

        return script

    pd.DataFrame._repr_javascript_ = _repr_datatable_

"""
import ipyvuetify as v
from traitlets import (Unicode, List)

def process_index(dataframe_, index_name="gene"):
    dataframe = dataframe_.copy()
    dataframe.index.name = index_name
    dataframe = dataframe.reset_index(drop=False)
    return dataframe

class voila_df(v.VuetifyTemplate):

    headers = List().tag(sync=True)
    items = List().tag(sync=True)
    template = Unicode().tag(sync=True)

    def __init__(self, df_):

        df = process_index(df_)

        super().__init__()
        columns = []
        v_slot = '<template v-slot:items="props">'
        for i, colname in enumerate(df.columns):
            if i == 0:
                list_item = {'text': colname, 'sortable': False, 'value': colname}
                v_slot = v_slot + f'<td>{{{{props.item.{colname}}}}}</td>'
            else:
                list_item = {'text': colname, 'value': colname, 'align': 'right'}
                v_slot = v_slot + f'<td class="text-xs-right">{{{{props.item.{colname}}}}}</td>'
            columns.append(list_item)
        v_slot = v_slot + '</template>'

        self.headers = columns
        self.items = df.to_dict(orient='records')
        self.template = '<v-data-table :headers="headers" :items="items" class="elevation-1">' \
            + v_slot \
            + '</v-data-table>'
"""



import os
import sys

import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns


import celloracle as co
print("co version: ", co.__version__)

plt.rcParams["figure.figsize"] = [6,6]
plt.rcParams["savefig.dpi"] = 600


def main():
    test_tutorial_simulation()
    
def test_tutorial_simulation():
    # Start analysis
    import time
    start = time.time()

    # Make folder to save plots
    save_folder = "figures"
    os.makedirs(save_folder, exist_ok=True)

    # 1. Load data
    print("\n1. Load data")
    oracle = co.data.load_tutorial_oracle_object()
    oracle
    # Here, we load demo links object for the training purpose.
    links = co.data.load_tutorial_links_object()

    # 2. Make predictive models for simulation
    print("\n2. Make predictive models for simulation")
    links.filter_links()
    oracle.get_cluster_specific_TFdict_from_Links(links_object=links)
    oracle.fit_GRN_for_simulation(alpha=10, use_cluster_specific_TFdict=True,
                                  verbose_level=1)

    # 3. In silico TF perturbation analysis
    print("\n3. In silico TF perturbation analysis")
    # Check gene expression
    goi = "Gata1"
    sc.pl.draw_graph(oracle.adata, color=[goi, oracle.cluster_column_name],
                     layer="imputed_count", use_raw=False, cmap="viridis")
    # Plot gene expression in histogram
    sc.get.obs_df(oracle.adata, keys=[goi], layer="imputed_count").hist()
    plt.show()
    # Enter perturbation conditions to simulate signal propagation after the perturbation.
    oracle.simulate_shift(perturb_condition={goi: 0.0},
                          n_propagation=3)
    # Get transition probability
    oracle.estimate_transition_prob(n_neighbors=200,
                                    knn_random=True,
                                    sampled_fraction=1)
    # Calculate embedding
    oracle.calculate_embedding_shift(sigma_corr=0.05)

    # 4. Visualization
    print("\n4. Visualization")
    fig, ax = plt.subplots(1, 2,  figsize=[13, 6])
    scale = 25
    # Show quiver plot
    oracle.plot_quiver(scale=scale, ax=ax[0])
    ax[0].set_title(f"Simulated cell identity shift vector: {goi} KO")
    # Show quiver plot that was calculated with randomized graph.
    oracle.plot_quiver_random(scale=scale, ax=ax[1])
    ax[1].set_title(f"Randomized simulation vector")
    plt.show()
    # Vector field graph visualization
    # n_grid = 40 is a good starting value.
    n_grid = 40
    oracle.calculate_p_mass(smooth=0.8, n_grid=n_grid, n_neighbors=200)
    # Search for best min_mass.
    oracle.suggest_mass_thresholds(n_suggestion=12)
    min_mass = 0.01
    oracle.calculate_mass_filter(min_mass=min_mass, plot=True)
    # Plot vector fields
    fig, ax = plt.subplots(1, 2,  figsize=[13, 6])
    scale_simulation = 0.5
    # Show quiver plot
    oracle.plot_simulation_flow_on_grid(scale=scale_simulation, ax=ax[0])
    ax[0].set_title(f"Simulated cell identity shift vector: {goi} KO")
    # Show quiver plot that was calculated with randomized graph.
    oracle.plot_simulation_flow_random_on_grid(scale=scale_simulation, ax=ax[1])
    ax[1].set_title(f"Randomized simulation vector")
    plt.show()
    # Plot vector field with cell cluster
    fig, ax = plt.subplots(figsize=[8, 8])
    oracle.plot_cluster_whole(ax=ax, s=10)
    oracle.plot_simulation_flow_on_grid(scale=scale_simulation,
                                        ax=ax, show_background=False)

    # 5. Compare simulation vector with development vectors
    print("\n5. Compare simulation vector with development vectors")
    # Prepare pseudotime data
    # Visualize pseudotime
    fig, ax = plt.subplots(figsize=[6,6])
    sc.pl.embedding(adata=oracle.adata, basis=oracle.embedding_name,
                    ax=ax, cmap="rainbow",
                    color=["Pseudotime"])
    # Calculation with Gradient calculator object
    from celloracle.applications import Gradient_calculator
    # Instantiate Gradient calculator object
    gradient = Gradient_calculator(oracle_object=oracle, pseudotime_key="Pseudotime")
    gradient.calculate_p_mass(smooth=0.8, n_grid=n_grid, n_neighbors=200)
    gradient.calculate_mass_filter(min_mass=min_mass, plot=True)
    gradient.transfer_data_into_grid(args={"method": "polynomial", "n_poly":3},
                                     plot=True)
    # Calculate graddient
    gradient.calculate_gradient()
    # Show results
    scale_dev = 40
    gradient.visualize_results(scale=scale_dev, s=5)
    # Visualize results
    fig, ax = plt.subplots(figsize=[6, 6])
    gradient.plot_dev_flow_on_grid(scale=scale_dev, ax=ax)
    # Save gradient object if you want.
    gradient.to_hdf5("Paul_etal.celloracle.gradient")
    # Calculate PS
    from celloracle.applications import Oracle_development_module
    # Make Oracle_development_module to compare two vector field
    dev = Oracle_development_module()
    # Load development flow
    dev.load_differentiation_reference_data(gradient_object=gradient)
    # Load simulation result
    dev.load_perturb_simulation_data(oracle_object=oracle)
    # Calculate inner produc scores
    dev.calculate_inner_product()
    dev.calculate_digitized_ip(n_bins=10)
    # Let's visualize the results
    dev.visualize_development_module_layout_0(s=5,
                                              scale_for_simulation=scale_simulation,
                                              s_grid=50,
                                              scale_for_pseudotime=scale_dev,
                                              vm=0.02)
    # Show perturbation scores
    fig, ax = plt.subplots(figsize=[6, 6])
    dev.plot_inner_product_on_grid(vm=0.02, s=50, ax=ax)
    # Show perturbation scores with perturbation simulation vector field
    fig, ax = plt.subplots(figsize=[6, 6])
    dev.plot_inner_product_on_grid(vm=0.02, s=50, ax=ax)
    dev.plot_simulation_flow_on_grid(scale=scale_simulation,
                                     show_background=False, ax=ax)

    # 6. Focus on a single development lineage to interpret the results in detail
    print("\n6. Focus on a single development lineage to interpret the results in detail")
    # Get cell index list for the cells of interest
    clusters = ['Ery_0', 'Ery_1', 'Ery_2', 'Ery_3', 'Ery_4', 'Ery_5', 'Ery_6',
                'Ery_7', 'Ery_8', 'Ery_9', 'MEP_0', 'Mk_0']
    cell_idx = np.where(oracle.adata.obs["louvain_annot"].isin(clusters))[0]
    # Check index
    print(cell_idx)
    # Calculate PS
    dev = Oracle_development_module()
    # Load development flow
    dev.load_differentiation_reference_data(gradient_object=gradient)

    # Load simulation result
    dev.load_perturb_simulation_data(oracle_object=oracle,
                                 cell_idx_use=cell_idx, # Enter cell id list
                                 name="Lineage_MEP" # Name of this cell group. You can enter any name.
                                 )
    # Calculation
    dev.calculate_inner_product()
    dev.calculate_digitized_ip(n_bins=10)
    # Let's visualize the results
    dev.visualize_development_module_layout_0(s=5,
                                              scale_for_simulation=scale_simulation,
                                              s_grid=50,
                                              scale_for_pseudotime=scale_dev,
                                              vm=0.03)
    # OLD VISUALIZATION
    from celloracle.visualizations.config import CONFIG
    CONFIG["cmap_ps"] = "coolwarm"
    dev.visualize_development_module_layout_0(s=5,
                                              scale_for_simulation=scale_simulation,
                                              s_grid=50,
                                              scale_for_pseudotime=scale_dev,
                                              vm=0.03)

    # Remove files
    print("Remove files")
    for i in ["Paul_etal.celloracle.gradient"]:
        os.remove(i)
        #pass
    os.removedirs("figures")
    elapsed = time.time() - start

    print("Success")
    print(f"Total time: {elapsed/60:.3g} minutes")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
'''

'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import pandas as pd
import numpy as np

import sys, os
import functools

#from tqdm.auto import tqdm
from tqdm.auto import tqdm
# 0.2. libraries for DNA and genome data wrangling and Motif analysis
from genomepy import Genome

#from gimmemotifs.motif import Motif
from gimmemotifs.scanner import Scanner
from gimmemotifs.fasta import Fasta
from gimmemotifs.config import DIRECT_NAME, INDIRECT_NAME
from gimmemotifs import __version__ as gmotif_version

from .modified_gimmemotif_function import custom_scan


from .process_bed_file import peak_M1

from .reference_genomes import SUPPORTED_REF_GENOME
from ..utility.package_version_checker import _is_version_OK

####
###

def get_available_ref_genome_info(species, provider="UCSC"):

    df = pd.DataFrame(np.array(list(genomepy.list_available_genomes(provider))))
    species_ = df[2].apply(lambda x: x.split(" ")[0])

    return df[species_.isin([species])]



def is_genome_installed(ref_genome):
    """
    Celloracle motif_analysis module uses gimmemotifs and genomepy internally.
    Reference genome files should be installed in the PC to use gimmemotifs and genomepy.
    This function checks the installation status of the reference genome.

    Args:
        ref_genome (str): names of reference genome. i.e., "mm10", "hg19"

    """
    try:
        genome_data = Genome(ref_genome)

        return True

    except:
        if ref_genome == "AmexG_v6.0-DD":
            print(f"genome {ref_genome} is not installed in this environment.")
            print(f"Please install {ref_genome} data with genomepy using the following command in command line (terminal).")
            print(f"Please make sure to use recent version of genomepy and have enough space on your PC. Otherwise you may get an error.")
            print(f'genomepy install -p url https://www.axolotl-omics.org/dl/AmexG_v6.0-DD.fa.gz')

        else:
            if  ref_genome in SUPPORTED_REF_GENOME.ref_genome.values:    
                provider = SUPPORTED_REF_GENOME["provider"][SUPPORTED_REF_GENOME.ref_genome==ref_genome].values[0]
            else:
                provider = "PROVIDER"

            print(f"genome {ref_genome} is not installed in this environment.")
            print("Please install genome using genomepy.")
            print(f'e.g.\n    >>> import genomepy\n    >>> genomepy.install_genome(name="{ref_genome}", provider="{provider}")')

    return False
        #raise ValueError(f"Ref_Genome: {ref_genome} is not available.")


def list2str(li):
    """
    Convert list of str into merged one str with " ,".
    See example below for detail.

    Args:
        li (list of str): list

    Returns:
        str: merged str.

    Examples:
        >>> a = ["a", "b", "c"]
        >>> list2str(a)
        "a, b, c"
    """
    return ", ".join(li)






def scan_dna_for_motifs(scanner_object, motifs_object, sequence_object, divide=100000, verbose=True):
    '''
    This is a wrapper function to scan DNA sequences searchig for Gene motifs.

    Args:

        scanner_object (gimmemotifs.scanner): Object that do motif scan.

        motifs_object (gimmemotifs.motifs): Object that stores motif data.

        sequence_object (gimmemotifs.fasta): Object that stores sequence data.

    Returns:
        pandas.dataframe: scan results is stored in data frame.

    '''

    li  = []
    # If the gimmemotifs version is larger than 0.17.0, it will automatically return the progress bar,
    # So we don't need to make tqdm qbar here.
    if _is_version_OK(que=gmotif_version, ref='0.17.0'):
        verbose=False

    pbar = tqdm(
        desc="scanning",
        unit=" sequences",
        total=len(sequence_object),
        disable=(verbose==False),  # can be silenced
    )
    iter = scanner_object.scan(sequence_object)

    for i, result in enumerate(iter):
        seqname = sequence_object.ids[i]
        for m,matches in enumerate(result):
            motif = motifs_object[m]
            for score, pos, strand in matches:
                li.append(np.array([seqname,
                                    motif.id,
                                    list2str(motif.factors[DIRECT_NAME]),
                                    list2str(motif.factors[INDIRECT_NAME]),
                                    score, pos, strand]))
        pbar.update(1)
    pbar.close()

    #save_as_pickled_object(li, "./tmp_li.pickle")
    #print("saved tmp list")


    if len(li)==0:
        df = pd.DataFrame(columns=["seqname",
                               "motif_id",
                               "factors_direct",
                               "factors_indirect",
                               "score", "pos", "strand"])
    else:

        remaining = 1
        LI = []
        k = 0
        while remaining == 1:
            #print(k)
            tmp_li = li[divide*k:min(len(li), divide*(k+1))]

            tmp_li = np.stack(tmp_li)
            df = pd.DataFrame(tmp_li,
                              columns=["seqname",
                                       "motif_id",
                                       "factors_direct",
                                       "factors_indirect",
                                       "score", "pos", "strand"])
            df.score = df.score.astype(float)
            df.pos = df.pos.astype(int)
            df.strand = df.strand.astype(int)

            # keep peak id name
            df.seqname = list(map(peak_M1, df.seqname.values))

            LI.append(df)


            if divide*(k+1) >= len(li):
                remaining = 0
            k += 1

        df = pd.concat(LI, axis=0)
    return df


# This is the preliminary version of mini-batch scan function. Keep this function just in case.
'''
def scan_dna_for_motifs_by_batch(scanner_object, motifs_object, sequence_object, batch_size=100, divide=100000, verbose=True):
    """
    This is a wrapper function to scan DNA sequences searchig for Gene motifs.

    Args:

        scanner_object (gimmemotifs.scanner): Object that do motif scan.

        motifs_object (gimmemotifs.motifs): Object that stores motif data.

        sequence_object (gimmemotifs.fasta): Object that stores sequence data.

    Returns:
        pandas.dataframe: scan results is stored in data frame.

    """
    li = scan_by_small_batch(sequence_object=sequence_object,
                             scanner_object=scanner_object,
                             motifs_object=motifs_object,
                             batch_size=batch_size,
                             verbose=verbose)



    if verbose:
        print("Motif Scan finished. Start post processing.")

    if len(li)==0:
        df = pd.DataFrame(columns=["seqname",
                               "motif_id",
                               "factors_direct",
                               "factors_indirect",
                               "score", "pos", "strand"])
    else:# Convert results as a dataframe little by little, and concat together later.

        remaining = 1
        LI = []
        k = 0
        while remaining == 1:
            #print(k)
            tmp_li = li[divide*k:min(len(li), divide*(k+1))]

            df = pd.DataFrame(tmp_li,
                              columns=["seqname",
                                       "motif_id",
                                       "factors_direct",
                                       "factors_indirect",
                                       "score", "pos", "strand"])
            df.score = df.score.astype(float)
            df.pos = df.pos.astype(int)
            df.strand = df.strand.astype(int)

            # keep peak id name
            df.seqname = list(map(peak_M1, df.seqname.values))

            LI.append(df)


            if divide*(k+1) >= len(li):
                remaining = 0
            k += 1

        df = pd.concat(LI, axis=0)
        df = df.reset_index(drop=True)

    return df

def _gimmemotifs_wrapper(sequence_object, scanner_object, motifs_object):
    li = []
    for i, result in enumerate(custom_scan(self=scanner_object, seqs=sequence_object, verbose=False)):
        seqname = sequence_object.ids[i]
        for m,matches in enumerate(result):
            motif = motifs_object[m]
            for score, pos, strand in matches:
                li.append([seqname,
                        motif.id,
                        _list2str(motif.factors[DIRECT_NAME]),
                        _list2str(motif.factors[INDIRECT_NAME]),
                                    score, pos, strand])
    return li

def scan_by_small_batch(sequence_object, scanner_object, motifs_object, batch_size=100, verbose=True):
    func = functools.partial(_gimmemotifs_wrapper, scanner_object=scanner_object, motifs_object=motifs_object)

    li = run_function_by_small_batch(function=func,
                                     input_data_iterable=sequence_object,
                                     batch_size=batch_size,
                                     return_method="add",
                                     verbose=verbose
                                     )
    return li

def run_function_by_small_batch(function=print, input_data_iterable=range(100), batch_size=30, return_method="list", verbose=True):
    """
    Run a function little by little.
    This can be applied to any function.
    This may be useful if a function can cause problem with large data.
    """
    batch_total = len(input_data_iterable) // batch_size
    if len(input_data_iterable) % batch_size != 0:
        batch_total += 1

    pbar = tqdm(
        desc="scanning",
        unit=" sequence_batch",
        total=batch_total,
        disable=(verbose==False),  # can be silenced
    )


    i = 0
    results = []
    while len(input_data_iterable) > batch_size*i:
        small_sequences = input_data_iterable[i*batch_size:batch_size*(i + 1)]
        out = function(small_sequences)
        if return_method == "add":
            results += out
        elif return_method == "list":
            results.append(out)
        pbar.update(1)
        i += 1
    pbar.close()

    return results
'''
